/*
    CSceneGraph: The Implementation (CAVE version)
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <cave_ogl.h>

#include "CSceneGraph.h"

// Constructor
CSceneGraph::CSceneGraph()
{
    // Initialize all attributes
    LastNode = NULL;
    NumOfNodes = 0;
    Nodes = NULL;
    NodePtr = NULL;
    SceneChanged = false;
    CounterID = 1;
}

// Destructor
CSceneGraph::~CSceneGraph()
{
    SceneGraph *temp, *ntemp;
    
    temp = Nodes;
    Nodes = NULL;
    SceneChanged = false;
    // Freeing all nodes available
    while(temp)
    {
        ntemp = temp->NextNode;
        CAVEFree(temp);
        temp = ntemp;
    }
    #ifdef _DEBUG_
        fprintf(stderr, "Object scene graph removed\n");
    #endif
}

// AddNode()
// Return value: 0 = NG else = OK
int CSceneGraph::AddNode(SceneGraph node, bool cnt)
{
    // In case of the first node...
    if(NumOfNodes == 0)
    {
        // Allocate and fill the node
        Nodes = (SceneGraph *) CAVEMalloc(sizeof(SceneGraph));
        bzero(Nodes, sizeof(SceneGraph));
        
        if(!Nodes)
        {
            #ifdef _DEBUG_
                fprintf(stderr, "Error: Cannot allocate memory\n");
            #endif
            return(0);
        }
        *Nodes = node;
        Nodes->NextNode = NULL;
        if(cnt)
        {
            Nodes->ObjID = CounterID;
            CounterID++;
        }
        NodePtr = Nodes;
        LastNode = Nodes;
        NumOfNodes++;
        // Flag on change...
        SceneChanged = true;
        return(Nodes->ObjID);
    }
    
    // Append as the last node
    LastNode->NextNode = (SceneGraph *) CAVEMalloc(sizeof(SceneGraph));
    bzero(LastNode->NextNode, sizeof(SceneGraph));
    
    if(!LastNode->NextNode)
    {
        #ifdef _DEBUG_
            fprintf(stderr, "Error: Cannot allocate memory\n");
        #endif
        return(0);
    }
    LastNode = LastNode->NextNode;
    *LastNode = node;
    LastNode->NextNode = NULL;
    if(cnt)
    {
        LastNode->ObjID = CounterID;
        CounterID++;
    }
    NumOfNodes++;
    // Flag on change...
    SceneChanged = true;
    return(LastNode->ObjID);
}

// DeleteNode()
// Return value: 0 = NG; 1 = OK
int CSceneGraph::DeleteNode(unsigned int index)
{
    SceneGraph *temp, *delnode;
    
    // If the node is not exist...
    if(index >= NumOfNodes)
    {
        #ifdef _DEBUG_
            fprintf(stderr, "Error: Cannot delete unexisting node\n");
        #endif
        return(0);
    }
    
    // Deleting the head of the list 
    if(index == 0)
    {
        delnode = Nodes;
        if(delnode == NodePtr)
            NodePtr = delnode->NextNode;
        if(delnode == LastNode);
            LastNode = NULL;
        Nodes = Nodes->NextNode;
        CAVEFree(delnode);
        NumOfNodes--;
        // Flag on change...
        SceneChanged = true;
        return(1);
    }
    
    // Deleting common nodes
    temp = Nodes;
    delnode = temp->NextNode;
    for(unsigned int i = 1; i < index; i++)
    {
        temp = temp->NextNode;
        delnode = temp->NextNode;
    }
    temp->NextNode = delnode->NextNode;
    if(delnode == NodePtr)
        NodePtr = temp->NextNode;
    if(delnode == LastNode)
        LastNode = temp;
    CAVEFree(delnode);
    NumOfNodes--;
    // Flag on change...
    SceneChanged = true;
    return(1);
}

// GetNodeAt()
// Return value: 0 = NG, 1 = OK
int CSceneGraph::GetNodeAt(unsigned int index, SceneGraph *node)
{
    SceneGraph *temp;
    
    // If the node is not exist
    if(index >= NumOfNodes)
    {
        #ifdef _DEBUG_
            fprintf(stderr, "Error: Cannot get unexisting node\n");
        #endif
        return(0);
    }
    
    // Get the node
    temp = Nodes;
    for(unsigned int i = 0; i < index; i++)
        if(temp)
            temp = temp->NextNode;
    
    if(!temp)
        return(0);
    
    if(!node)
    {
        node = (SceneGraph *) CAVEMalloc(sizeof(SceneGraph));
        bzero(node, sizeof(SceneGraph));
        if(!node)
        {
            #ifdef _DEBUG_
                fprintf(stderr, "Error: Cannot allocate memory\n");
            #endif
            return(0);
        }
    }
    
    *node = *temp;
    node->NextNode = NULL;
    
    return(1);
}

// DeleteNodeWithID()
// Return value: 0 = NG; 1 = OK
int CSceneGraph::DeleteNodeWithID(GLuint theID)
{
    SceneGraph *temp, *delnode;
    
    if(NumOfNodes == 0)
        return(0);

    if(Nodes->ObjID == theID)
    {
        temp = Nodes;
        if(LastNode == temp)
            LastNode = NULL;
        Nodes = Nodes->NextNode;
        CAVEFree(temp);
        NumOfNodes--;
        // Flag on change...
        SceneChanged = true;
        return(1);
    }
    
    delnode = Nodes->NextNode;
    temp = Nodes;
    
    do
    {
        if(delnode->ObjID == theID)
        {
            temp->NextNode = delnode->NextNode;
            if(LastNode == delnode)
                LastNode = temp;
            CAVEFree(delnode);
            NumOfNodes--;
            // Flag on change...
            SceneChanged = true;
            return(1);
        }
        else
        {
            temp = temp->NextNode;
            delnode = delnode->NextNode;
        }
    }
    while(delnode);
    
    return(0);
}

// GetNodeWithID()
// Return value: 0 = NG, 1 = OK
int CSceneGraph::GetNodeWithID(GLuint theID, SceneGraph *node)
{
    SceneGraph *ptr;
    if(NumOfNodes == 0)
        return(0);
    
    ptr = Nodes;
    
    do
    {
        if(ptr->ObjID == theID)
        {
            if(!node)
            {
                node = (SceneGraph *) CAVEMalloc(sizeof(SceneGraph));
                bzero(node, sizeof(SceneGraph));
            }
            *node = *ptr;
            return(1);
        }
        else
            ptr = ptr->NextNode;
    }
    while(ptr);
    
    return(0);
}

// RewindNodePointer()
void CSceneGraph::RewindNodePointer(void)
{
    NodePtr = Nodes;
}

// GetCurrentNodePointer()
// Return value: -1 = NG; 0 = No other nodes; 1 = Other node is available
int CSceneGraph::GetCurrentNodePointer(SceneGraph *node)
{
    if(!NodePtr)
        return(0);

    if(!node)
    {
        node = (SceneGraph *) CAVEMalloc(sizeof(SceneGraph));
        bzero(node, sizeof(SceneGraph));
        if(!node)
        {
            #ifdef _DEBUG_
                fprintf(stderr, "Error: Cannot allocate memory\n");
            #endif
            return(-1);
        }
    }
    
    // Get current node pointed by NodePtr
    *node = *NodePtr;
    node->NextNode = NULL;
    
    // Advance node to the next node of the list
    NodePtr = NodePtr->NextNode;

    if(NodePtr)
        return(1);
    else
        return(0);
}

// UpdateNode()
// Return value: 0 = NG; 1 = OK
int CSceneGraph::UpdateNode(SceneGraph node)
{
    SceneGraph *temp;
    
    temp = Nodes;
    
    while(temp)
    {
        if(temp->ObjID == node.ObjID)
        {
            *temp = node;
            // Flag on change...
            SceneChanged = true;
            return(1);
        }
        else
            temp = temp->NextNode;
    }
    return(0);
}

// Serialize()
// Return value: 0 = NG; 1 = OK
int CSceneGraph::Serialize(const char *filename)
{
    FILE *nusfile = NULL;
    SceneGraph *temp;
    int lock = -1;
    
    // Open file
    if((nusfile = fopen(filename, "wb")) == NULL)
    {
        #ifdef _DEBUG_
            fprintf(stderr, "Error: Cannot open file...\n");
        #endif
        return(0);
    }
    
    // First, write number of entities available and counter...
    fwrite(&NumOfNodes, sizeof(unsigned int), 1, nusfile);
    fwrite(&CounterID, sizeof(GLuint), 1, nusfile);
    
    // Write all nodes as binaries, all nodes must be lock-free
    temp = Nodes;
    while(temp)
    {
        // Save locking state
        if(temp->Locked != -1)
        {
            lock = temp->Locked;
            temp->Locked = -1;
        }
        if(fwrite(temp, sizeof(SceneGraph), 1, nusfile) != 1)
        {
            #ifdef _DEBUG_
                fprintf(stderr, "Error: Cannot write to file...\n");
            #endif
            return(0);
        }
        // Restore locking state
        if(lock != -1)
        {
            temp->Locked = lock;
            lock = -1;
        }
        temp = temp->NextNode;
    }
    
    fclose(nusfile);
    // Flag on change...
    SceneChanged = false;
    return(1);
}

// LoadFromFile()
// Return value: 0 = NG; 1 = OK
// Warning: all nodes will be deleted and replaced by nodes from file.
int CSceneGraph::LoadFromFile(const char *filename)
{
    FILE *nusfile = NULL;
    SceneGraph currnode;
    unsigned int temp;
    
    // Flag on change...
    SceneChanged = false;
    // Load list from files
    if((nusfile = fopen(filename, "rb")) == NULL)
    {
        #ifdef _DEBUG_
            fprintf(stderr, "Error: Cannot open file...\n");
        #endif
        return(0);
    }

    // if there are already nodes in the list, delete it... all!
    DeleteAllNodes();
    
    // Read first entity from the file, number of nodes...
    if(fread(&temp, sizeof(unsigned int), 1, nusfile) != 1)
    {
        #ifdef _DEBUG_
            fprintf(stderr, "Error: File corrupted...\n");
        #endif
        return(0);
    }
    
    if(fread(&CounterID, sizeof(GLuint), 1, nusfile) != 1)
    {
        #ifdef _DEBUG_
            fprintf(stderr, "Error: File corrupted...\n");
        #endif
        return(0);
    }
    
    while(!feof(nusfile))
    {
        // Read and add to scene graph
        fread(&currnode, sizeof(SceneGraph), 1, nusfile);
        if(!feof(nusfile))
            if(AddNode(currnode, false) == 0)
            {
                #ifdef _DEBUG_
                    fprintf(stderr, "Error: Can't add node...\n");
                #endif
                DeleteAllNodes();
                return(0);
            }
    }
    
    if(NumOfNodes != temp)
    {
        #ifdef _DEBUG_
            fprintf(stderr, "Error: Incomplete read...\n");
        #endif
        DeleteAllNodes();
        return(0);
    }
    
    fclose(nusfile);
    return(1);
}

// DeleteAllNodes()
void CSceneGraph::DeleteAllNodes(void)
{
    SceneGraph *temp, *ntemp;
    
    temp = Nodes;
    Nodes = NULL;
    // Delete each node
    while(temp)
    {
        ntemp = temp->NextNode;
        CAVEFree(temp);
        temp = ntemp;
    }
    
    LastNode = NULL;
    NodePtr = NULL;
    NumOfNodes = 0;
    CounterID = 1;
}

// NumberOfNodes()
unsigned int CSceneGraph::NumberOfNodes(void)
{
    return NumOfNodes;
}
