/*
    CAVE Stand-alone Client
    Created by Nus, 2003
    Description: Stand-alone client for CAVE
*/

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <GL/gl.h>
#include <GL/glx.h>
#include <cave_ogl.h>
#include <math.h>

#include "objfile.h"
#include "CSceneGraph.h"
#include "CObjReference.h"
#include "CImmersiveWorld.h"
#include "stdfunc.h"
#include "MotionData.h"
#include "Scine.h"

CSceneGraph *scene = NULL;
CAVELOCK datalock;
CImmersiveWorld *world = NULL;
CAVELOCK worldlock;

MotionData *motion = NULL;
CAVELOCK motiondata;

CObjReference *objref = NULL;
//GLuint hand_obj;
GLuint font_list;

void InitGL(char *reference);
void InitSharedMemory(char *scenefile);
void RenderScene(void);
void FrameUpdate(void);

void RotateVec(float degree[3], float in[3], float out[3]);
float AngleCosValue(float A[3], float B[3]);
bool ObjectTouching(SceneGraph scn, Object3D obj, float point[3]);
void SelectObject(void);
void PrintString(char *str);

void * operator new(size_t, void *location);

float VectorProc(float x1,float y1,float z1,float x2,float y2,float z2);
bool HandTouching();

void main(int argc, char **argv)
{
    CAVESetOption(CAVE_SHMEM_SIZE, 67108864);
    CAVEConfigure(&argc, argv, NULL);
    
    if(argc != 4)
    {
        printf("Usage: caveclient [CAVE options] c|s source ref_dir\n");
        CAVEExit();
    }
    
    if(argv[1][0] == 's')
    {
        InitSharedMemory(argv[2]);
        CAVEInit();
        CAVEInitApplication((CAVECALLBACK) InitGL, 1, argv[3]);
    
        CAVEDisplay((CAVECALLBACK) RenderScene, 0);
        CAVEFrameFunction((CAVECALLBACK) FrameUpdate, 0);
    }
    else
    {
        printf("Not yet available...\n");
        CAVEExit();
    }
    
    while(!CAVEgetbutton(CAVE_ESCKEY))
    {
		  //algorithm();
        sginap(10);
    }
    
    scene->~CSceneGraph();
    CAVEFree(scene);
    
    world->~CImmersiveWorld();
    CAVEFree(world);
    
	 motion->~MotionData();
	 CAVEFree(motion); 

//    glDeleteLists(hand_obj, 1);
    glDeleteLists(font_list, 256);
    
    CAVEExit();
}

void InitGL(char *reference)
{
    GLfloat ltAmb[] = {0.5f, 0.5f, 0.5f, 1.0f};
    GLfloat ltPos[] = {0.0f, 50.0f, 0.0f, 1.0f};
//    Object3D temp_obj;
    
    glClearColor(0.0, 0.0, 0.0, 1.0);
    glShadeModel(GL_SMOOTH);
    glEnable(GL_DEPTH_TEST);
    glEnable(GL_CULL_FACE);
    glEnable(GL_RESCALE_NORMAL);
    glFrontFace(GL_CCW);
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    
    glEnable(GL_LIGHTING);
    glLightModelfv(GL_LIGHT_MODEL_AMBIENT, ltAmb);
    glLightfv(GL_LIGHT0, GL_POSITION, ltPos);
    glEnable(GL_LIGHT0);
    
    glLineWidth(5.0f);
    
    font_list = glGenLists(256);
    glXUseXFont(XLoadFont(CAVEXDisplay(),
        "-adobe-helvetica-bold-r-normal--*-200-*-*-p-*-iso8859-1"), 0, 256,
        font_list);
    
    //hand_obj = ReadObject("hand.obj", &temp_obj);
    
    if((objref = new CObjReference()) == NULL)
    {
        printf("CAVE Error: Error allocating memory for database...\n");
        printf("CAVE Info: Broadcasting exit message...\n");
        CAVEExit();
    }
    
    if((objref->OpenLibrary(reference)) == 0)
    {
        printf("CAVE Error: Error reading database...\n");
        printf("CAVE Info: Broadcasting exit message...\n");
        CAVEExit();
    }
    else
        printf("CAVE Info: Succesfully loading %d object(s) from %s\n",
            objref->NumberOfObjects(), reference);
}

void InitSharedMemory(char *scenefile)
{
    CSceneGraph *temp = NULL;
    CImmersiveWorld *tempwrl = NULL;
	 MotionData *tempmot = NULL;
    
    temp = (CSceneGraph *) CAVEMalloc(sizeof(CSceneGraph));
    if(temp == NULL)
    {
        printf("CAVE Error: Error initializing shared memory for scene\n");
        printf("CAVE Info: Broadcasting exit message...\n");
        CAVEExit();
    }
    else
    {
        scene = new (temp) CSceneGraph();
        scene->LoadFromFile(scenefile);
        printf("CAVE Info: Succesfully loading %d object(s)\n",
            scene->NumberOfNodes());
    }
    datalock = CAVENewLock();
    
    tempwrl = (CImmersiveWorld *) CAVEMalloc(sizeof(CImmersiveWorld));
    if(tempwrl == NULL)
    {
        printf("CAVE Error: Error initializing shared memory for world\n");
        printf("CAVE Info:Broadcasting exit message...\n");
        CAVEExit();
    }
    else
        world = new (tempwrl) CImmersiveWorld();
    worldlock = CAVENewLock();

    tempmot = (MotionData*) CAVEMalloc(sizeof(MotionData));
    if(tempmot == NULL)
    {
        printf("CAVE Error: Error initializing shared memory for world\n");
        printf("CAVE Info:Broadcasting exit message...\n");
        CAVEExit();
    }
    else
        motion = new (tempmot) MotionData();
    motiondata = CAVENewLock();
}

void RenderScene(void)
{

    int result;
    SceneGraph now;
    Object3D theObject;
    float nav_trans[3], nav_rot;
	 float nav_scale;
    GLuint selected;

    char nav_state;
    //char axis_dep;

	 float nav_sen[3];
    
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    
	 //SCINE
	 scine();
    CAVESetReadLock(worldlock);
    
    world->GetNavTranslation(nav_trans);
    nav_scale = world->GetNavScale();
    nav_rot = world->GetNavRotation();
    selected = world->selected;

    nav_state = world->GetNavState();
    //axis_dep = world->GetAxisDep();

    CAVEUnsetReadLock(worldlock);

    CAVESetReadLock(motiondata);
	 motion->GetNavTranslation(nav_sen);
    CAVEUnsetReadLock(motiondata);


    glEnable(GL_COLOR_MATERIAL);

    glColor3f(1.0f, 1.0f, 1.0f);

    if(CAVEWall == CAVE_FRONT_WALL)
    {
        char buff[64];
        
        glMatrixMode(GL_PROJECTION);
        glPushMatrix();
        glMatrixMode(GL_MODELVIEW);
        glPushMatrix();
        CAVEWallTransform();
        glDisable(GL_LIGHTING);

        glRasterPos2f(0.3f, 7.7f);

        sprintf(buff, "Selected: %d", selected);
        PrintString(buff);
        
/*
        glRasterPos2f(0.3f, 7.3f);

        switch(axis_dep)
        {
        case AXIS_DEP_X:
            PrintString("Axis Dep.: X");
            break;
        case AXIS_DEP_Y:
            PrintString("Axis Dep.: Y");
            break;
        case AXIS_DEP_Z:
            PrintString("Axis Dep.: Z");
            break;
        case AXIS_DEP_XY:
            PrintString("Axis Dep.: XY");
            break;
        case AXIS_DEP_YZ:
            PrintString("Axis Dep.: YZ");
            break;
        case AXIS_DEP_XZ:
            PrintString("Axis Dep.: XZ");
            break;
        case AXIS_DEP_XYZ:
            PrintString("Axis Dep.: XYZ");
        }
*/

        glRasterPos2f(0.3f, 6.9f);

        switch(nav_state)
        {
        case NAV_WALK:
            PrintString("Walk Mode");
            break;
        case NAV_FLOAT:
            PrintString("Float Mode");
            break;
        case NAV_SCALE:
            PrintString("Scale Mode");
            break;
/*
        case OBJ_SELECT:
            PrintString("Select Object");
            break;
        case OBJ_TRANSLATE:
            PrintString("Translate Object");
            break;
        case OBJ_ROTATE:
            PrintString("Rotate Object");
            break;
        case OBJ_SCALE:
            PrintString("Scale Object");
*/
        }

        glEnable(GL_LIGHTING);
        glMatrixMode(GL_MODELVIEW);
        glPopMatrix();
        glMatrixMode(GL_PROJECTION);
        glPopMatrix();
        glMatrixMode(GL_MODELVIEW);
    }
    
/*
    glPushMatrix();
        CAVEWandTransform();
        glTranslatef(0.0f, 0.0f, -0.3f);
        glFrontFace(GL_CW);
        glCallList(hand_obj);
        glFrontFace(GL_CCW);
    glPopMatrix();
*/


    glDisable(GL_COLOR_MATERIAL);

    CAVENavLoadIdentity();

    //CAVENavWorldScale(nav_scale, nav_scale, nav_scale);
    CAVENavScale(nav_scale, nav_scale, nav_scale);

    CAVENavTranslate(nav_trans[0], nav_trans[1], nav_trans[2]);
    //CAVENavWorldTranslate(nav_trans[0], nav_trans[1], nav_trans[2]);
    CAVENavTranslate(nav_sen[0], nav_sen[1], nav_sen[2]);
	 //motion->GetPrint();
    CAVENavRot(nav_rot, 'y');
    CAVENavTransform();
    
/*
   glTranslatef( CAVESensor10 -> x,  CAVESensor10 -> y, CAVESensor10 -> z);
   glRotatef(CAVESensor10->azim, 0.0f,1.0f,0.0f);
   glRotatef(CAVESensor10->elev, 1.0f,0.0f,0.0f);
   glRotatef(CAVESensor10->roll, 0.0f,0.0f,1.0f);
   glTranslatef( -CAVESensor10 -> x,  -CAVESensor10 -> y, -CAVESensor10 -> z);
*/

    CAVESetWriteLock(datalock);
    
    if(scene->NumberOfNodes() > 0)
    {
        scene->RewindNodePointer();
        do
        {
            result = scene->GetCurrentNodePointer(&now);
            
		   float sen10hand[3];
		   CAVEGetSensorPosition(CAVESENSOR(10),CAVE_TRACKER_FRAME,sen10hand);

            glPushMatrix();
/*
   			    glTranslatef( sen10hand[0],  sen10hand[1], sen10hand[2]);
					 glRotatef(now.ObjRotation.x, 0.0f, 1.0f, 0.0f);
                glRotatef(now.ObjRotation.y, 1.0f, 0.0f, 0.0f);
                glRotatef(now.ObjRotation.z, 0.0f, 0.0f, 1.0f);
   	          glTranslatef( -sen10hand[0],  -sen10hand[1], -sen10hand[2]);
*/

                glTranslatef(now.ObjLocation.x, now.ObjLocation.y,
                    now.ObjLocation.z);
//ZYX YZX YXZ
                glRotatef(now.ObjRotation.y, 0.0f, 1.0f, 0.0f);
                glRotatef(now.ObjRotation.x, 1.0f, 0.0f, 0.0f);
                glRotatef(now.ObjRotation.z, 0.0f, 0.0f, 1.0f);
/*
                glRotatef(now.ObjRotation.x, 0.0f, 1.0f, 0.0f);
                glRotatef(now.ObjRotation.y, 0.0f, 0.0f, 1.0f);
                glRotatef(now.ObjRotation.z, 0.0f, 1.0f, 0.0f);
*/
                glScalef(now.ObjScaling.x, now.ObjScaling.y, now.ObjScaling.z);
                objref->ReferObjectAt(now.ReferenceNum, &theObject);
                glCallList(theObject.DisplayListNum);
                
                if(selected != 0 && now.ObjID == selected)
                {
                    glEnable(GL_COLOR_MATERIAL);
                    glDisable(GL_LIGHTING);
                    glColor3f(1.0f, 0.0f, 0.0f);
                    glBegin(GL_LINE_LOOP);
                        glVertex3f(theObject.MaximumBoundingBox.x,
                            theObject.MaximumBoundingBox.y,
                            theObject.MaximumBoundingBox.z);
                        glVertex3f(theObject.MinimumBoundingBox.x,
                            theObject.MaximumBoundingBox.y,
                            theObject.MaximumBoundingBox.z);
                        glVertex3f(theObject.MinimumBoundingBox.x,
                            theObject.MinimumBoundingBox.y,
                            theObject.MaximumBoundingBox.z);
                        glVertex3f(theObject.MaximumBoundingBox.x,
                            theObject.MinimumBoundingBox.y,
                            theObject.MaximumBoundingBox.z);
                        glVertex3f(theObject.MaximumBoundingBox.x,
                            theObject.MinimumBoundingBox.y,
                            theObject.MinimumBoundingBox.z);
                        glVertex3f(theObject.MinimumBoundingBox.x,
                            theObject.MinimumBoundingBox.y,
                            theObject.MinimumBoundingBox.z);
                        glVertex3f(theObject.MinimumBoundingBox.x,
                            theObject.MaximumBoundingBox.y,
                            theObject.MinimumBoundingBox.z);
                        glVertex3f(theObject.MaximumBoundingBox.x,
                            theObject.MaximumBoundingBox.y,
                            theObject.MinimumBoundingBox.z);
                    glEnd();
                    glBegin(GL_LINES);
                        glVertex3f(theObject.MaximumBoundingBox.x,
                            theObject.MaximumBoundingBox.y,
                            theObject.MaximumBoundingBox.z);
                        glVertex3f(theObject.MaximumBoundingBox.x,
                            theObject.MinimumBoundingBox.y,
                            theObject.MaximumBoundingBox.z);
                        glVertex3f(theObject.MaximumBoundingBox.x,
                            theObject.MaximumBoundingBox.y,
                            theObject.MinimumBoundingBox.z);
                        glVertex3f(theObject.MaximumBoundingBox.x,
                            theObject.MinimumBoundingBox.y,
                            theObject.MinimumBoundingBox.z);
                        glVertex3f(theObject.MinimumBoundingBox.x,
                            theObject.MaximumBoundingBox.y,
                            theObject.MaximumBoundingBox.z);
                        glVertex3f(theObject.MinimumBoundingBox.x,
                            theObject.MaximumBoundingBox.y,
                            theObject.MinimumBoundingBox.z);
                        glVertex3f(theObject.MinimumBoundingBox.x,
                            theObject.MinimumBoundingBox.y,
                            theObject.MaximumBoundingBox.z);
                        glVertex3f(theObject.MinimumBoundingBox.x,
                            theObject.MinimumBoundingBox.y,
                            theObject.MinimumBoundingBox.z);
                    glEnd();
                    glDisable(GL_COLOR_MATERIAL);
                    glEnable(GL_LIGHTING);
                }
            glPopMatrix();
        }
        while(result == 1);
    }
    
    CAVEUnsetWriteLock(datalock);
}

static int flag = 0;
//static float presen[3];

void FrameUpdate(void)
{

    if(CAVEMasterDisplay())
    {
        //char nav_state;
        //char axis_dep;
        //float joyx, joyy;
        GLuint selected;
        
        CAVESetWriteLock(worldlock);
        world->DoTransform();
        CAVEUnsetWriteLock(worldlock);

        CAVESetWriteLock(motiondata);
	     motion->DoTransform();
        CAVEUnsetWriteLock(motiondata);
        
        if(CAVEButtonChange(3) == -1)
        {
            CAVESetWriteLock(worldlock);
            world->ToggleState();
            CAVEUnsetWriteLock(worldlock);
        }

/*
        if(CAVEButtonChange(2) == -1)
        {
            CAVESetWriteLock(worldlock);
            world->IterateAxisDep();
            CAVEUnsetWriteLock(worldlock);
        }
*/


		  HandTouching();

        //if(CAVEButtonChange(1) == 1)
        if(motion->checkhand == 1 && flag != 5)
        {

			flag = 5;
			
		  //printf("motion->checkhand = %d \n", motion->checkhand);
		  //printf("world->selected= %d \n", world->selected);



        //    char mode;
            CAVESetWriteLock(worldlock);
        //    mode = world->GetNavState();
            world->selected = 0;
            CAVEUnsetWriteLock(worldlock);
				
            //if(mode == OBJ_SELECT || mode == OBJ_TRANSLATE ||
            //    mode == OBJ_ROTATE || mode == OBJ_SCALE)
				
            SelectObject();
        }

        //if(CAVEButtonChange(3) == -1){
        if(motion->checkhand == 0){
		  //printf("motion->checkhand = %d \n", motion->checkhand);
		  //printf("world->selected= %d \n", world->selected);
            world->selected = 0;
				motion->anglehand = 0;
			   flag = 0;
		  }

        CAVESetReadLock(worldlock);
        //nav_state = world->GetNavState();
        //axis_dep = world->GetAxisDep();
        selected = world->selected;
        CAVEUnsetReadLock(worldlock);

        //if(selected != 0 && (nav_state == OBJ_TRANSLATE ||
        //    nav_state == OBJ_ROTATE || nav_state == OBJ_SCALE))

        if(selected != 0)
        {
/*
	      float           wandPos[3],wandOri[3];
  	   	float           wandWorldPos[3];
      	CAVEGetPosition(CAVE_WAND, wandPos);
      	CAVEGetOrientation(CAVE_WAND, wandOri);
      	CAVENavConvertCAVEToWorld(wandPos, wandWorldPos);
*/

		   float sen10Pos[3],sen10Ori[3],sen10WorldPos[3];
//		   float sen10Pos[3],sen10WorldPos[3];
//		   float sen10Vec[3],sen10WorldVec[3];
		   CAVEGetSensorPosition(CAVESENSOR(10),CAVE_TRACKER_FRAME,sen10Pos);
		   CAVENavConvertCAVEToWorld(sen10Pos, sen10WorldPos);

//		   CAVEGetSensorVector(CAVESENSOR(10),CAVE_TRACKER_FRAME,sen10Vec);
//		   CAVENavConvertVectorCAVEToWorld(sen10Vec, sen10WorldVec);

		   //CAVEGetSensorOrientation(CAVESENSOR(10),CAVE_TRACKER_FRAME,sen10Ori);
		   CAVEGetSensorOrientation(CAVESENSOR(10),CAVE_NAV_FRAME,sen10Ori);

/*
   glTranslatef( CAVESensor10 -> x,  CAVESensor10 -> y, CAVESensor10 -> z);
   glRotatef(CAVESensor10->azim, 0.0f,1.0f,0.0f);
   glRotatef(CAVESensor10->elev, 1.0f,0.0f,0.0f);
   glRotatef(CAVESensor10->roll, 0.0f,0.0f,1.0f);
   glTranslatef( -CAVESensor10 -> x,  -CAVESensor10 -> y, -CAVESensor10 -> z);
*/

/*
			CAVENavWorldRot(sen10Ori[0], 'x');
			CAVENavWorldRot(sen10Ori[1], 'y');
			CAVENavWorldRot(sen10Ori[2], 'z');
*/

	/*
			CAVEGetSensorOrientation();
			float wandOri[3];
			CAVEGetOrientation(CAVE_WAND,wandOri);
			CAVENavWorldRot(float angle, char axis);
	*/ 


       //    joyx = CAVE_JOYSTICK_X;
        //   joyy = CAVE_JOYSTICK_Y;

       //     if(fabs(joyx) > 0.2f || fabs(joyy) > 0.2f)
       //     {
                SceneGraph temp;

       //         switch(nav_state)
       //         {
       //         case OBJ_TRANSLATE:
       //             switch(axis_dep)
       //             {
       //             case AXIS_DEP_X:

								//static float tempangle[3];
							
// y-x-z z-y-x 
                        CAVESetWriteLock(datalock);
                        scene->GetNodeWithID(selected, &temp);

                        temp.ObjLocation.x = sen10WorldPos[0];
                        temp.ObjLocation.y = sen10WorldPos[1];
                        temp.ObjLocation.z = sen10WorldPos[2];
                        temp.ObjRotation.x = sen10Ori[0]; 
                        temp.ObjRotation.y = sen10Ori[1];
                        temp.ObjRotation.z = sen10Ori[2];
/*
								if(motion->anglehand == 0){
								temp.ObjRotation.x += 0;
								temp.ObjRotation.y += 0;
								temp.ObjRotation.z += 0;
								presen[0] = sen10Ori[0];
								presen[1] = sen10Ori[1];
								presen[2] = sen10Ori[2];

//								tempangle[0] = sen10Ori[0] - temp.ObjRotation.x;
//								tempangle[1] = sen10Ori[1] - temp.ObjRotation.y;
//								tempangle[2] = sen10Ori[2] - temp.ObjRotation.z;
//								printf("HANDHAND");
//								printf("%f",tempangle[0] = sen10Ori[0] - temp.ObjRotation.x);

								motion->anglehand = 1;
								}
								temp.ObjRotation.x += sen10Ori[0] - presen[0];
								temp.ObjRotation.y += sen10Ori[1] - presen[1];
								temp.ObjRotation.z += sen10Ori[2] - presen[2];
								presen[0] = sen10Ori[0];
								presen[1] = sen10Ori[1];
								presen[2] = sen10Ori[2];
*/

/*
								printf("Obj = %f \n",temp.ObjRotation.x);
                        temp.ObjRotation.x = sen10Ori[0] - tempangle[0]; 
                        temp.ObjRotation.y = sen10Ori[1] - tempangle[1];
                        temp.ObjRotation.z = sen10Ori[2] - tempangle[2];
								printf("Obj2 = %f \n",temp.ObjRotation.x);
                        temp.ObjRotation.x = sen10Ori[0]; 
                        temp.ObjRotation.y = sen10Ori[1];
                        temp.ObjRotation.z = sen10Ori[2];
*/

                        scene->UpdateNode(temp);
                        CAVEUnsetWriteLock(datalock);
/*
                        CAVESetWriteLock(datalock);
                        scene->GetNodeWithID(selected, &temp);
                        temp.ObjLocation.x = wandWorldPos[0];
                        temp.ObjLocation.y = wandWorldPos[1];
                        temp.ObjLocation.z = wandWorldPos[2];
                        temp.ObjRotation.x = wandOri[0];
                        temp.ObjRotation.y = wandOri[1];
                        temp.ObjRotation.z = wandOri[2];
                        scene->UpdateNode(temp);
                        CAVEUnsetWriteLock(datalock);
*/

        /*                break;
                    case AXIS_DEP_Y:
                        CAVESetWriteLock(datalock);
                        scene->GetNodeWithID(selected, &temp);
                        temp.ObjLocation.y += joyy * 0.05f;
                        scene->UpdateNode(temp);
                        CAVEUnsetWriteLock(datalock);
                        break;
                    case AXIS_DEP_Z:
                        CAVESetWriteLock(datalock);
                        scene->GetNodeWithID(selected, &temp);
                        temp.ObjLocation.z -= joyy * 0.05f;
                        scene->UpdateNode(temp);
                        CAVEUnsetWriteLock(datalock);
                        break;
                    case AXIS_DEP_XY:
                        CAVESetWriteLock(datalock);
                        scene->GetNodeWithID(selected, &temp);
                        temp.ObjLocation.x += joyx * 0.05f;
                        temp.ObjLocation.y += joyy * 0.05f;
                        scene->UpdateNode(temp);
                        CAVEUnsetWriteLock(datalock);
                        break;
                    case AXIS_DEP_YZ:
                        CAVESetWriteLock(datalock);
                        scene->GetNodeWithID(selected, &temp);
                        temp.ObjLocation.y += joyx * 0.05f;
                        temp.ObjLocation.z += joyy * 0.05f;
                        scene->UpdateNode(temp);
                        CAVEUnsetWriteLock(datalock);
                        break;
                    case AXIS_DEP_XZ:
                        CAVESetWriteLock(datalock);
                        scene->GetNodeWithID(selected, &temp);
                        temp.ObjLocation.x += joyx * 0.05f;
                        temp.ObjLocation.z -= joyy * 0.05f;
                        scene->UpdateNode(temp);
                        CAVEUnsetWriteLock(datalock);
                    }
                    break;
                case OBJ_ROTATE:
                    switch(axis_dep)
                    {
                    case AXIS_DEP_X:
                        CAVESetWriteLock(datalock);
                        scene->GetNodeWithID(selected, &temp);
                        temp.ObjRotation.x -= joyy * 0.1f;
                        scene->UpdateNode(temp);
                        CAVEUnsetWriteLock(datalock);
                        break;
                    case AXIS_DEP_Y:
                        CAVESetWriteLock(datalock);
                        scene->GetNodeWithID(selected, &temp);
                        temp.ObjRotation.y += joyx * 0.1f;
                        scene->UpdateNode(temp);
                        CAVEUnsetWriteLock(datalock);
                        break;
                    case AXIS_DEP_Z:
                        CAVESetWriteLock(datalock);
                        scene->GetNodeWithID(selected, &temp);
                        temp.ObjRotation.z -= joyx * 0.1f;
                        scene->UpdateNode(temp);
                        CAVEUnsetWriteLock(datalock);
                        break;
                    case AXIS_DEP_XY:
                        CAVESetWriteLock(datalock);
                        scene->GetNodeWithID(selected, &temp);
                        temp.ObjRotation.x -= joyy * 0.1f;
                        temp.ObjRotation.y += joyx * 0.1f;
                        scene->UpdateNode(temp);
                        CAVEUnsetWriteLock(datalock);
                        break;
                    case AXIS_DEP_YZ:
                        CAVESetWriteLock(datalock);
                        scene->GetNodeWithID(selected, &temp);
                        temp.ObjRotation.y += joyx * 0.1f;
                        temp.ObjRotation.z -= joyy * 0.1f;
                        scene->UpdateNode(temp);
                        CAVEUnsetWriteLock(datalock);
                        break;
                    case AXIS_DEP_XZ:
                        CAVESetWriteLock(datalock);
                        scene->GetNodeWithID(selected, &temp);
                        temp.ObjRotation.x -= joyy * 0.1f;
                        temp.ObjRotation.z -= joyx * 0.1f;
                        scene->UpdateNode(temp);
                        CAVEUnsetWriteLock(datalock);
                    }
                    break;
                case OBJ_SCALE:
                    switch(axis_dep)
                    {
                    case AXIS_DEP_X:
                        CAVESetWriteLock(datalock);
                        scene->GetNodeWithID(selected, &temp);
                        temp.ObjScaling.x += joyy * 0.05f;
                        scene->UpdateNode(temp);
                        CAVEUnsetWriteLock(datalock);
                        break;
                    case AXIS_DEP_Y:
                        CAVESetWriteLock(datalock);
                        scene->GetNodeWithID(selected, &temp);
                        temp.ObjScaling.y += joyy * 0.05f;
                        scene->UpdateNode(temp);
                        CAVEUnsetWriteLock(datalock);
                        break;
                    case AXIS_DEP_Z:
                        CAVESetWriteLock(datalock);
                        scene->GetNodeWithID(selected, &temp);
                        temp.ObjScaling.z += joyy * 0.05f;
                        scene->UpdateNode(temp);
                        CAVEUnsetWriteLock(datalock);
                        break;
                    case AXIS_DEP_XY:
                        CAVESetWriteLock(datalock);
                        scene->GetNodeWithID(selected, &temp);
                        temp.ObjScaling.x += joyx * 0.05f;
                        temp.ObjScaling.y += joyy * 0.05f;
                        scene->UpdateNode(temp);
                        CAVEUnsetWriteLock(datalock);
                        break;
                    case AXIS_DEP_YZ:
                        CAVESetWriteLock(datalock);
                        scene->GetNodeWithID(selected, &temp);
                        temp.ObjScaling.y += joyx * 0.05f;
                        temp.ObjScaling.z += joyy * 0.05f;
                        scene->UpdateNode(temp);
                        CAVEUnsetWriteLock(datalock);
                        break;
                    case AXIS_DEP_XZ:
                        CAVESetWriteLock(datalock);
                        scene->GetNodeWithID(selected, &temp);
                        temp.ObjScaling.x += joyx * 0.05f;
                        temp.ObjScaling.z += joyy * 0.05f;
                        scene->UpdateNode(temp);
                        CAVEUnsetWriteLock(datalock);
                        break;
                    case AXIS_DEP_XYZ:
                        CAVESetWriteLock(datalock);
                        scene->GetNodeWithID(selected, &temp);
                        temp.ObjScaling.x += joyy * 0.05f;
                        temp.ObjScaling.y += joyy * 0.05f;
                        temp.ObjScaling.z += joyy * 0.05f;
                        scene->UpdateNode(temp);
                        CAVEUnsetWriteLock(datalock);
                    }
                    break;
                }
            }
								*/
        }
    }
}

void RotateVec(float degree[3], float in[3], float out[3])
{
    float RotMat[3][3];
    float tr[3];
    float sphi, cphi;
    float stheta, ctheta;
    float spsi, cpsi;

    sphi = sinf(degree[0] / 180.0f * M_PI);
    cphi = cosf(degree[0] / 180.0f * M_PI);
    stheta = sinf(degree[1] / 180.0f * M_PI);
    ctheta = cosf(degree[1] / 180.0f * M_PI);
    spsi = sinf(degree[2] / 180.0f * M_PI);
    cpsi = cosf(degree[2] / 180.0f * M_PI);

    RotMat[0][0] = cpsi*ctheta*cphi - spsi*sphi;
    RotMat[0][1] = cpsi*stheta;
    RotMat[0][2] = -cpsi*ctheta*sphi - spsi*cphi;
    RotMat[1][0] = -stheta*cphi;
    RotMat[1][1] = ctheta;
    RotMat[1][2] = stheta*sphi;
    RotMat[2][0] = spsi*ctheta*cphi + cpsi*sphi;
    RotMat[2][1] = spsi*stheta;
    RotMat[2][2] = -spsi*ctheta*sphi + cpsi*cphi;

    tr[0] = in[0]*RotMat[0][0] + in[1]*RotMat[1][0] + in[2]*RotMat[2][0];
    tr[1] = in[0]*RotMat[0][1] + in[1]*RotMat[1][1] + in[2]*RotMat[2][1];
    tr[2] = in[0]*RotMat[0][2] + in[1]*RotMat[1][2] + in[2]*RotMat[2][2];

    out[0] = tr[0];
    out[1] = tr[1];
    out[2] = tr[2];
}

float AngleCosValue(float A[3], float B[3])
{
    return((A[0]*B[0] + A[1]*B[1] + A[2]*B[2]) /
        sqrtf(A[0]*A[0] + A[1]*A[1] + A[2]*A[2]) /
        sqrtf(B[0]*B[0] + B[1]*B[1] + B[2]*B[2]));
}

bool ObjectTouching(SceneGraph scn, Object3D obj, float point[3])
{
    float v1[3], v2[3], v3[3];
    float nMax[3], nMin[3];
    float pMax[3], pMin[3];
    float testv[3];
    float rot[3];
    
    v1[0] = obj.MinimumBoundingBox.x - obj.MaximumBoundingBox.x;
    v1[1] = 0.0f;
    v1[2] = 0.0f;
    v2[0] = 0.0f;
    v2[1] = obj.MinimumBoundingBox.y - obj.MaximumBoundingBox.y;
    v2[2] = 0.0f;
    v3[0] = 0.0f;
    v3[1] = 0.0f;
    v3[2] = obj.MinimumBoundingBox.z - obj.MaximumBoundingBox.z;
    
    nMax[0] = scn.ObjScaling.x * obj.MaximumBoundingBox.x;
    nMax[1] = scn.ObjScaling.y * obj.MaximumBoundingBox.y;
    nMax[2] = scn.ObjScaling.z * obj.MaximumBoundingBox.z;

    nMin[0] = scn.ObjScaling.x * obj.MinimumBoundingBox.x;
    nMin[1] = scn.ObjScaling.y * obj.MinimumBoundingBox.y;
    nMin[2] = scn.ObjScaling.z * obj.MinimumBoundingBox.z;
    
    rot[0] = scn.ObjRotation.x;
    rot[1] = scn.ObjRotation.y;
    rot[2] = scn.ObjRotation.z;
    
    RotateVec(rot, nMax, nMax);
    RotateVec(rot, nMin, nMin);
    RotateVec(rot, v1, v1);
    RotateVec(rot, v2, v2);
    RotateVec(rot, v3, v3);
    
    pMax[0] = scn.ObjLocation.x + nMax[0];
    pMax[1] = scn.ObjLocation.y + nMax[1];
    pMax[2] = scn.ObjLocation.z + nMax[2];

    pMin[0] = scn.ObjLocation.x + nMin[0];
    pMin[1] = scn.ObjLocation.y + nMin[1];
    pMin[2] = scn.ObjLocation.z + nMin[2];
    
    testv[0] = point[0] - pMax[0];
    testv[1] = point[1] - pMax[1];
    testv[2] = point[2] - pMax[2];
    
    if(AngleCosValue(testv, v1) <= 0.0f)
        return(false);
    if(AngleCosValue(testv, v2) <= 0.0f)
        return(false);
    if(AngleCosValue(testv, v3) <= 0.0f)
        return(false);
    
    v1[0] = -v1[0];
    v1[1] = -v1[1];
    v1[2] = -v1[2];
    v2[0] = -v2[0];
    v2[1] = -v2[1];
    v2[2] = -v2[2];
    v3[0] = -v3[0];
    v3[1] = -v3[1];
    v3[2] = -v3[2];
    
    testv[0] = point[0] - pMin[0];
    testv[1] = point[1] - pMin[1];
    testv[2] = point[2] - pMin[2];

    if(AngleCosValue(testv, v1) <= 0.0f)
        return(false);
    if(AngleCosValue(testv, v2) <= 0.0f)
        return(false);
    if(AngleCosValue(testv, v3) <= 0.0f)
        return(false);

    return(true);
}

void SelectObject(void)
{
    //float point[3], vector[3];

    float point[3];
    SceneGraph now;
    Object3D theObject;
    int result;
    
   float sen10[3],sen10WorldPos[3];
   CAVEGetSensorPosition(CAVESENSOR(10),CAVE_TRACKER_FRAME,sen10);
   CAVENavConvertCAVEToWorld(sen10, sen10WorldPos);
   point[0] = sen10WorldPos[0]; 
   point[1] = sen10WorldPos[1]; 
   point[2] = sen10WorldPos[2]; 

/*
    CAVEGetPosition(CAVE_WAND_NAV, point);
    CAVEGetVector(CAVE_WAND_FRONT_NAV, vector);
    point[0] += 0.75 * vector[0];
    point[1] += 0.75 * vector[1];
    point[2] += 0.75 * vector[2];
    CAVEGetVector(CAVE_WAND_LEFT_NAV, vector);
    point[0] += 0.1 * vector[0];
    point[1] += 0.1 * vector[1];
    point[2] += 0.1 * vector[2];
*/
    
    CAVESetWriteLock(datalock);

    if(scene->NumberOfNodes() > 0)
    {
        scene->RewindNodePointer();
        do
        {
            result = scene->GetCurrentNodePointer(&now);
            objref->ReferObjectAt(now.ReferenceNum, &theObject);
            if(ObjectTouching(now, theObject, point))
            {
                CAVESetWriteLock(worldlock);
                world->selected = now.ObjID;
                CAVEUnsetWriteLock(worldlock);
                break;
            }
        }
        while(result == 1);
    }
    
    CAVEUnsetWriteLock(datalock);
}

void PrintString(char *str)
{
    glPushAttrib(GL_LIST_BIT);
        glListBase(font_list);
        glCallLists(strlen(str), GL_UNSIGNED_BYTE, (GLubyte *) str);
    glPopAttrib();
}

void * operator new(size_t, void *location)
{
    return(location);
}

float VectorProc(float x1,float y1,float z1,float x2,float y2,float z2)
{
      return sqrt((x1-x2)*(x1-x2)+(y1-y2)*(y1-y2)+(z1-z2)*(z1-z2));
}

bool HandTouching()
{
	float vec;
	
   float sen10[3],sen11[3];
   float sen10WorldPos[3],sen11WorldPos[3];
   CAVEGetSensorPosition(CAVESENSOR(10),CAVE_TRACKER_FRAME,sen10);
   CAVEGetSensorPosition(CAVESENSOR(11),CAVE_TRACKER_FRAME,sen11);
   CAVENavConvertCAVEToWorld(sen10, sen10WorldPos);
   CAVENavConvertCAVEToWorld(sen11, sen11WorldPos);

	vec = VectorProc(sen10WorldPos[0],sen10WorldPos[1],sen10WorldPos[2],
		              sen11WorldPos[0],sen11WorldPos[1],sen11WorldPos[2]);

	//printf("sen10[0] = %f \n sen11[0] = %f \n",sen10WorldPos[0],sen11WorldPos[0]);
	//printf("Vector = %f \n ", vec);

	//Long 0.19 hand 0.10  then sclae 0.3f
	//Long 0.77 hand 0.328 
	//if(vec < 0.14){
	if(vec < 0.5){
	motion->checkhand = 1;
	return true;
	}else{
	motion->checkhand = 0;
	return false;
	}

}

