/*
    CAVE Client's Standard Functions Implementation
    Program by Nus
*/
#include <math.h>

#include "stdfunc.h"

void RotateVec(float degree[3], float in[3], float out[3])
{
    float RotMat[3][3];
    float tr[3];
    float sphi, cphi;
    float stheta, ctheta;
    float spsi, cpsi;
    
    sphi = sinf(degree[0] / 180.0f * M_PI);
    cphi = cosf(degree[0] / 180.0f * M_PI);
    stheta = sinf(degree[1] / 180.0f * M_PI);
    ctheta = cosf(degree[1] / 180.0f * M_PI);
    spsi = sinf(degree[2] / 180.0f * M_PI);
    cpsi = cosf(degree[2] / 180.0f * M_PI);

    RotMat[0][0] = cpsi*ctheta*cphi - spsi*sphi;
    RotMat[0][1] = cpsi*stheta;
    RotMat[0][2] = -cpsi*ctheta*sphi - spsi*cphi;
    RotMat[1][0] = -stheta*cphi;
    RotMat[1][1] = ctheta;
    RotMat[1][2] = stheta*sphi;
    RotMat[2][0] = spsi*ctheta*cphi + cpsi*sphi;
    RotMat[2][1] = spsi*stheta;
    RotMat[2][2] = -spsi*ctheta*sphi + cpsi*cphi;
    
    tr[0] = in[0]*RotMat[0][0] + in[1]*RotMat[1][0] + in[2]*RotMat[2][0];
    tr[1] = in[0]*RotMat[0][1] + in[1]*RotMat[1][1] + in[2]*RotMat[2][1];
    tr[2] = in[0]*RotMat[0][2] + in[1]*RotMat[1][2] + in[2]*RotMat[2][2];
    
    out[0] = tr[0];
    out[1] = tr[1];
    out[2] = tr[2];
}

void * operator new(size_t, void *location)
{
    return(location);
}

float AngleCosValue(float A[3], float B[3])
{
    return((A[0]*B[0] + A[1]*B[1] + A[2]*B[2]) /
        sqrtf(A[0]*A[0] + A[1]*A[1] + A[2]*A[2]) /
        sqrtf(B[0]*B[0] + B[1]*B[1] + B[2]*B[2]));
}

bool ObjectTouching(SceneGraph scn, Object3D obj, float point[3])
{
    float v1[3], v2[3], v3[3];
    float nMax[3], nMin[3];
    float pMax[3], pMin[3];
    float testv[3];
    float rot[3];
    
    v1[0] = obj.MinimumBoundingBox.x - obj.MaximumBoundingBox.x;
    v1[1] = 0.0f;
    v1[2] = 0.0f;
    v2[0] = 0.0f;
    v2[1] = obj.MinimumBoundingBox.y - obj.MaximumBoundingBox.y;
    v2[2] = 0.0f;
    v3[0] = 0.0f;
    v3[1] = 0.0f;
    v3[2] = obj.MinimumBoundingBox.z - obj.MaximumBoundingBox.z;
    
    nMax[0] = scn.ObjScaling.x * obj.MaximumBoundingBox.x;
    nMax[1] = scn.ObjScaling.y * obj.MaximumBoundingBox.y;
    nMax[2] = scn.ObjScaling.z * obj.MaximumBoundingBox.z;

    nMin[0] = scn.ObjScaling.x * obj.MinimumBoundingBox.x;
    nMin[1] = scn.ObjScaling.y * obj.MinimumBoundingBox.y;
    nMin[2] = scn.ObjScaling.z * obj.MinimumBoundingBox.z;
    
    rot[0] = scn.ObjRotation.x;
    rot[1] = scn.ObjRotation.y;
    rot[2] = scn.ObjRotation.z;
    
    RotateVec(rot, nMax, nMax);
    RotateVec(rot, nMin, nMin);
    RotateVec(rot, v1, v1);
    RotateVec(rot, v2, v2);
    RotateVec(rot, v3, v3);
    
    pMax[0] = scn.ObjLocation.x + nMax[0];
    pMax[1] = scn.ObjLocation.y + nMax[1];
    pMax[2] = scn.ObjLocation.z + nMax[2];

    pMin[0] = scn.ObjLocation.x + nMin[0];
    pMin[1] = scn.ObjLocation.y + nMin[1];
    pMin[2] = scn.ObjLocation.z + nMin[2];
    
    testv[0] = point[0] - pMax[0];
    testv[1] = point[1] - pMax[1];
    testv[2] = point[2] - pMax[2];
    
    if(AngleCosValue(testv, v1) <= 0.0f)
        return(false);
    if(AngleCosValue(testv, v2) <= 0.0f)
        return(false);
    if(AngleCosValue(testv, v3) <= 0.0f)
        return(false);
    
    v1[0] = -v1[0];
    v1[1] = -v1[1];
    v1[2] = -v1[2];
    v2[0] = -v2[0];
    v2[1] = -v2[1];
    v2[2] = -v2[2];
    v3[0] = -v3[0];
    v3[1] = -v3[1];
    v3[2] = -v3[2];
    
    testv[0] = point[0] - pMin[0];
    testv[1] = point[1] - pMin[1];
    testv[2] = point[2] - pMin[2];

    if(AngleCosValue(testv, v1) <= 0.0f)
        return(false);
    if(AngleCosValue(testv, v2) <= 0.0f)
        return(false);
    if(AngleCosValue(testv, v3) <= 0.0f)
        return(false);

    return(true);
}
